/*! @file advanced-sensing/camera-h264-hardware-decode-sample.cpp
 *  @version 3.4
 *  @date Dec 17 2020
 *
 *  @brief
 *  Logging API usage in a Linux environment.
 *  Shows example usage of various advanced sensing APIs and controls.
 *
 *  @Copyright (c) 2017 DJI
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <dji_vehicle.hpp>
#include <dji_linux_helpers.hpp>
#include "camera_h264_hardware_decoder.hpp"
#include "camera_stream_show.hpp"

using namespace DJI;
using namespace DJI::OSDK;

CameraStreamShow *s_cameraStreamShow = nullptr;

void cameraH264DataCb(uint8_t *buf, int bufLen, void *userData)
{
	ErrorCode::ErrorCodeType errorCodeType;
	auto *s_cameraH264HardwareDecoder = (CameraH264HardwareDecoder *) userData;

	errorCodeType = s_cameraH264HardwareDecoder->DecodeH264StreamData(buf, bufLen);
	if (errorCodeType != ErrorCode::SUCCESS) {
		DERROR("Decode H264 Stream error, errCode = %lld", errorCodeType);
	}
}

void cameraRgbImageCallback(const T_Image *image)
{
    s_cameraStreamShow->ShowImage(image);
}

int main(int argc, char **argv)
{
	// Setup OSDK.
	PayloadIndexType payloadIndex;
	bool enableAdvancedSensing = true;
	bool quitSample = false;
	LiveView::LiveViewCameraPosition cameraPosition;
	LinuxSetup linuxEnvironment(argc, argv, enableAdvancedSensing);
	Vehicle *vehicle = linuxEnvironment.getVehicle();
	if (vehicle == nullptr) {
		std::cout << "Vehicle not initialized, exiting.\n";
		return -1;
	}
	CameraH264HardwareDecoder *s_cameraH264HardwareDecoder;

	/* Initialize GStreamer */
	gst_init(nullptr, nullptr);

	/*! sample loop start */
	char inputChar = 0;

	std::cout << std::endl;
	std::cout
		<< "| Available commands:                                             |"
		<< std::endl
		<< "| [a] Start main camera H264 hardware decode sample               |"
		<< std::endl
		<< "| [b] Start vice camera H264 hardware decode sample               |"
		<< std::endl
		<< "| [c] Start top camera H264 hardware decode sample                |"
		<< std::endl
		<< "| [q] Stop camera H264 hardware decode sample                     |"
		<< std::endl;

	std::cin >> inputChar;

	switch (inputChar) {
		case 'a':
			payloadIndex = PAYLOAD_INDEX_0;
			cameraPosition = DJI::OSDK::LiveView::OSDK_CAMERA_POSITION_NO_1;
			break;
		case 'b':
			payloadIndex = PAYLOAD_INDEX_1;
			cameraPosition = DJI::OSDK::LiveView::OSDK_CAMERA_POSITION_NO_2;
			break;
		case 'c':
			payloadIndex = PAYLOAD_INDEX_2;
			cameraPosition = DJI::OSDK::LiveView::OSDK_CAMERA_POSITION_NO_3;
			break;
		default:
			std::cout << "Unavailable commands [" << inputChar << "], Please input correct char!" << endl;
			exit(1);
	}

	s_cameraH264HardwareDecoder = new CameraH264HardwareDecoder(vehicle, payloadIndex);
	s_cameraH264HardwareDecoder->RegisterRgbImageCallback(cameraRgbImageCallback);
	s_cameraStreamShow = new CameraStreamShow();

	vehicle->advancedSensing->startH264Stream(cameraPosition,
	                                          cameraH264DataCb,
	                                          s_cameraH264HardwareDecoder);
	while (!quitSample) {
		std::cin >> inputChar;
		if (inputChar == 'q') {
			quitSample = true;
		}
	}
	vehicle->advancedSensing->stopH264Stream(cameraPosition);

	return 0;
}
