/**
 ********************************************************************
 * @file    camera_h264_hardware_decoder.cpp.c
 * @brief
 *
 * @copyright (c) 2018 DJI. All rights reserved.
 *
 * All information contained herein is, and remains, the property of DJI.
 * The intellectual and technical concepts contained herein are proprietary
 * to DJI and may be covered by U.S. and foreign patents, patents in process,
 * and protected by trade secret or copyright law.  Dissemination of this
 * information, including but not limited to data and other proprietary
 * material(s) incorporated within the information, in any form, is strictly
 * prohibited without the express written consent of DJI.
 *
 * If you receive this source code without DJI’s authorization, you may not
 * further disseminate the information, and you must immediately remove the
 * source code and notify DJI of its removal. DJI reserves the right to pursue
 * legal actions against you for any loss(es) or damage(s) caused by your
 * failure to do so.
 *
 *********************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include <dji_vehicle.hpp>
#include <pthread.h>
#include "camera_h264_hardware_decoder.hpp"
#include <gst/app/gstappsrc.h>
#include <gst/app/gstappsink.h>
#include <gst/gstbuffer.h>

/* Private constants ---------------------------------------------------------*/

/* Private types -------------------------------------------------------------*/
typedef struct {
	const char *cameraTypeStr;
	CameraModule::WorkMode workMode;

	IdrFrameType idrFrameType;
} T_CameraVideoStreamResolutionTableItem;

/* Private values -------------------------------------------------------------*/
static T_CameraVideoStreamResolutionTableItem s_cameraVideoStreamResolutionTable[] = {
	{"H20", CameraModule::SHOOT_PHOTO,  IDR_FRAME_TYPE_H20_SHOT_PHOTO},
	{"H20", CameraModule::RECORD_VIDEO, IDR_FRAME_TYPE_H20_RECORD_VIDEO},
	{"Z30", CameraModule::SHOOT_PHOTO,  IDR_FRAME_TYPE_Z30_SHOT_PHOTO},
	{"Z30", CameraModule::RECORD_VIDEO, IDR_FRAME_TYPE_Z30_RECORD_VIDEO},
};

/* Private functions declaration ---------------------------------------------*/
void* UpdateRealIdrFrameTask(void *arg);
void* GetNewFrameTask(void *arg);

/* Exported functions definition ---------------------------------------------*/
CameraH264HardwareDecoder::CameraH264HardwareDecoder(Vehicle *vehicle,
                                                     PayloadIndexType payloadIndex)
{
	m_vehicle = vehicle;
	m_payloadIndex = payloadIndex;
	m_exitUpdateRealIdrFrameTypeTask = false;
	m_exitGetNewFrameTask = false;
	m_decoderUseIdrFrameType = IDR_FRAME_TYPE_UNKNOWN;
	m_realIdrFrameType = IDR_FRAME_TYPE_UNKNOWN;
	m_imageCallback = nullptr;

	/* Create Gst pipeline */
	if (CreateGstDecodePipeline() != ErrorCode::SUCCESS) {
		DERROR("Create gst decode pipeline error");
		exit(1);
	}
	
	//init camera module
	if (m_vehicle->cameraManager->initCameraModule(payloadIndex, "camera") != ErrorCode::SUCCESS) {
		DERROR("Init camera module error, please make sure camera is mounted");
		exit(1);
	}

	//create get frame resolution thread
	OsdkOsal_TaskCreate(&m_updateRealIdrFrameTypeTaskHandle, UpdateRealIdrFrameTask, 4096, this);
}

CameraH264HardwareDecoder::~CameraH264HardwareDecoder()
{
	m_exitUpdateRealIdrFrameTypeTask = true;
	OsdkOsal_TaskDestroy(m_updateRealIdrFrameTypeTaskHandle);

	DestoryGstDecodePipeline();
}

ErrorCode::ErrorCodeType CameraH264HardwareDecoder::CreateGstDecodePipeline()
{
	GstStateChangeReturn ret;

	/* Create the elements */
	m_gstAppSrc = gst_element_factory_make("appsrc", "dji_video_stream_src");
	m_gstQueue = gst_element_factory_make("queue", "queue");
	m_gstH264Parse = gst_element_factory_make("h264parse", "h264parse");
	m_gstDecode = gst_element_factory_make("omxh264dec", "decode");
	m_gstAppSink = gst_element_factory_make("appsink", "sink");

	/* Create the empty pipeline */
	m_gstPipeline = gst_pipeline_new("dji_video_stream_decode_pipeline");

	if (!m_gstPipeline || !m_gstAppSrc || !m_gstQueue || !m_gstH264Parse || !m_gstDecode ||
	    !m_gstAppSink) {
		DERROR("Not all gstreamer elements could be created.");
		return ErrorCode::SYSTEM_ERROR;
	}

	/* Build the pipeline */
	gst_bin_add_many(GST_BIN (m_gstPipeline), m_gstAppSrc, m_gstQueue, m_gstH264Parse, m_gstDecode,
	                 m_gstAppSink, NULL);
	if (gst_element_link_many(m_gstAppSrc, m_gstQueue, m_gstH264Parse, m_gstDecode, m_gstAppSink,
	                          NULL) != TRUE) {
		DERROR("Gstreamer elements could not be linked.");
		gst_object_unref(m_gstPipeline);
		return ErrorCode::SYSTEM_ERROR;
	}

	//appsrc config
	g_object_set(m_gstAppSrc,
    			"stream-type", GST_APP_STREAM_TYPE_STREAM,
    			"format", GST_FORMAT_TIME,
    			"do-timestamp", TRUE,
    			"is-live", TRUE,
    			"block", FALSE,
    			NULL);

	//decode disable dpb for low lantency
	g_object_set(m_gstDecode, "disable-dpb", TRUE, NULL);

    // appsink config
	/* max queue buffer numbers */
	gst_app_sink_set_max_buffers(GST_APP_SINK(m_gstAppSink), 10);
	/* drop old buffers when queue is filled */
	gst_app_sink_set_drop(GST_APP_SINK(m_gstAppSink), true);

	/* Start playing */
	ret = gst_element_set_state(m_gstPipeline, GST_STATE_PLAYING);
	if (ret == GST_STATE_CHANGE_FAILURE) {
		DERROR("Unable to set the pipeline to the playing state.");
		gst_object_unref(m_gstPipeline);
		return ErrorCode::SYSTEM_ERROR;
	}
	// create get frame task
	m_exitGetNewFrameTask = false;
	m_IsGetNewFrameTaskLive = false;
	OsdkOsal_TaskCreate(&m_getNewFrameTaskHandle, GetNewFrameTask, 4096, this);

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraH264HardwareDecoder::DestoryGstDecodePipeline()
{
    GstStateChangeReturn stateRet;
    GstState currentState;
    GstState currentPending;

	//exit get new frame task
	m_exitGetNewFrameTask = true;
    gst_element_send_event (m_gstPipeline, gst_event_new_eos());
	while (m_IsGetNewFrameTaskLive) {
		usleep(1000);
	}
	OsdkOsal_TaskDestroy(m_getNewFrameTaskHandle);

	//reset pipeline
    gst_element_set_state(m_gstPipeline, GST_STATE_PAUSED);
	gst_object_unref(m_gstPipeline);

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraH264HardwareDecoder::RestartGstPipeline()
{
	ErrorCode::ErrorCodeType errorCodeType;

	errorCodeType = DestoryGstDecodePipeline();
	if (errorCodeType != ErrorCode::SUCCESS) {
		DERROR("Destroy gst decode pipeline error");
		return errorCodeType;
	}

	errorCodeType = CreateGstDecodePipeline();
	if (errorCodeType != ErrorCode::SUCCESS) {
		DERROR("Create gst decode pipeline error");
		return errorCodeType;
	}

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraH264HardwareDecoder::DecodeH264StreamData(const uint8_t *data, uint32_t dataLen)
{
	ErrorCode::ErrorCodeType errorCodeType;
	const uint8_t *idrFrameDataPtr;
	uint32_t idrFrameDataLen;
	IdrFrameType currenIdrFrameType;

	if (m_realIdrFrameType == IDR_FRAME_TYPE_UNKNOWN) {
		//unknown, can't decode, wait get frame resolution
		return ErrorCode::SUCCESS;
	}

	currenIdrFrameType = m_realIdrFrameType;

	if (m_decoderUseIdrFrameType != currenIdrFrameType) {
	    DSTATUS("Restart pipeline");
	    if (RestartGstPipeline() != ErrorCode::SUCCESS) {
	    	DERROR("Restart pipeline failed");
			return ErrorCode::SYSTEM_ERROR;
		}

		//Need push idr frame to hardware decoder to identify h264 stream, because dji camera h264 steam use GDR
		errorCodeType = H264IdrFrame_GetData(currenIdrFrameType, idrFrameDataPtr, idrFrameDataLen);
		if (errorCodeType != ErrorCode::SUCCESS) {
			DERROR("Can't get IDR Frame, resolution enum value = %lld.", currenIdrFrameType);
			return errorCodeType;
		}
		errorCodeType = DecodeDataUseDecoder(idrFrameDataPtr, idrFrameDataLen);
		if (errorCodeType != ErrorCode::SUCCESS) {
			DERROR("Hardware decoder decode data error, errorCode = %lld", errorCodeType);
			return errorCodeType;
		}
		m_decoderUseIdrFrameType = currenIdrFrameType;
	} else {
		errorCodeType = DecodeDataUseDecoder(data, dataLen);
		if (errorCodeType != ErrorCode::SUCCESS) {
			DERROR("Hardware decoder decode data error, errorCode = %lld", errorCodeType);
			return errorCodeType;
		}
	}

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraH264HardwareDecoder::RegisterRgbImageCallback(ImageCallback imageCallback)
{
	if (imageCallback == nullptr) {
		return ErrorCode::PARAMETER_SET_FAILED;
	}
	m_imageCallback = imageCallback;

	return ErrorCode::SUCCESS;
}

/* Private functions definition-----------------------------------------------*/
ErrorCode::ErrorCodeType CameraH264HardwareDecoder::DecodeDataUseDecoder(const uint8_t *data, uint32_t dataLen)
{
	GstBuffer *gstDecodeBuffer;
	GstFlowReturn ret;

	 gstDecodeBuffer = gst_buffer_new_allocate(nullptr, dataLen, nullptr);
	 gst_buffer_fill(gstDecodeBuffer, 0, data, dataLen);

	 ret = gst_app_src_push_buffer(GST_APP_SRC(m_gstAppSrc), gstDecodeBuffer);
	 if (ret != GST_FLOW_OK) {
	 	DERROR("Gstreamer appSrc push buffer error!");
	 }

	return ErrorCode::SUCCESS;
}

void* UpdateRealIdrFrameTask(void *arg)
{
	CameraH264HardwareDecoder *s_cameraH264HardwareDecoder = (CameraH264HardwareDecoder *)arg;
	int i;
	ErrorCode::ErrorCodeType errorCode;
	CameraModule::WorkMode workMode;
	string cameraVersion;

	while (!s_cameraH264HardwareDecoder->m_exitUpdateRealIdrFrameTypeTask) {
		errorCode = s_cameraH264HardwareDecoder->m_vehicle->cameraManager->getModeSync(
			s_cameraH264HardwareDecoder->m_payloadIndex, workMode, 1000);
		if (errorCode != ErrorCode::SUCCESS) {
			DERROR("Get camera mode error!");
			continue;
		}
		cameraVersion = s_cameraH264HardwareDecoder->m_vehicle->cameraManager->getCameraVersion(
			s_cameraH264HardwareDecoder->m_payloadIndex);

		for (i = 0;
		     i < sizeof(s_cameraVideoStreamResolutionTable) / sizeof(T_CameraVideoStreamResolutionTableItem); i++) {
			if ((strcmp(s_cameraVideoStreamResolutionTable[i].cameraTypeStr, cameraVersion.c_str()) == 0) &&
			    s_cameraVideoStreamResolutionTable[i].workMode == workMode) {
				s_cameraH264HardwareDecoder->m_realIdrFrameType = s_cameraVideoStreamResolutionTable[i].idrFrameType;
				break;
			}
		}
		if (i == sizeof(s_cameraVideoStreamResolutionTable) / sizeof(T_CameraVideoStreamResolutionTableItem)) {
			DERROR("Can't find this frame resolution on this type camera, hardware decode stop");
			s_cameraH264HardwareDecoder->m_realIdrFrameType = IDR_FRAME_TYPE_UNKNOWN;
		}

		//update frame resolution every 500ms
		usleep(500000);
	}
}

void* GetNewFrameTask(void *arg)
{
	CameraH264HardwareDecoder *s_cameraH264HardwareDecoder = (CameraH264HardwareDecoder *)arg;
	GstStateChangeReturn ret;
	T_Image image;
	GstSample *gstImageSample;
	GstBuffer *gstImageBuffer;
	GstCaps *caps;
	GstStructure *gstStructure;
	const char *format;
	GstMapInfo gstMap;

	s_cameraH264HardwareDecoder->m_IsGetNewFrameTaskLive = true;

	while (!s_cameraH264HardwareDecoder->m_exitGetNewFrameTask) {
		gstImageSample = gst_app_sink_pull_sample(GST_APP_SINK(s_cameraH264HardwareDecoder->m_gstAppSink));
		if (gstImageSample != nullptr) {
			gstImageBuffer = gst_sample_get_buffer(gstImageSample);
			caps = gst_sample_get_caps(gstImageSample);
			gstStructure = gst_caps_get_structure(caps, 0);

			if (gst_buffer_map (gstImageBuffer, &gstMap, GST_MAP_READ)){
				image.width = g_value_get_int(gst_structure_get_value(gstStructure, "width"));
				image.height = g_value_get_int(gst_structure_get_value(gstStructure, "height"));
				format = g_value_get_string(gst_structure_get_value(gstStructure, "format"));
				strcpy(image.format, format);
				image.imageData = gstMap.data;
				image.imageDataLen = gstMap.size;
				 if (s_cameraH264HardwareDecoder->m_imageCallback != nullptr) {
				 	s_cameraH264HardwareDecoder->m_imageCallback(&image);
				 }
				gst_buffer_unmap(gstImageBuffer, &gstMap);
			} else {
				DERROR("Gst buffer can't map");
			}
            gst_sample_unref(gstImageSample);
		} else {
			if (gst_app_sink_is_eos(GST_APP_SINK(s_cameraH264HardwareDecoder->m_gstAppSink))) {
                DSTATUS("Get EOS for appsrc");
			}
		}
		usleep(1000);
	}

	s_cameraH264HardwareDecoder->m_IsGetNewFrameTaskLive = false;

	return nullptr;
}

/****************** (C) COPYRIGHT DJI Innovations *****END OF FILE****/
