/**
 ********************************************************************
 * @file    camera_stream_show.cpp
 * @brief
 *
 * @copyright (c) 2018 DJI. All rights reserved.
 *
 * All information contained herein is, and remains, the property of DJI.
 * The intellectual and technical concepts contained herein are proprietary
 * to DJI and may be covered by U.S. and foreign patents, patents in process,
 * and protected by trade secret or copyright law.  Dissemination of this
 * information, including but not limited to data and other proprietary
 * material(s) incorporated within the information, in any form, is strictly
 * prohibited without the express written consent of DJI.
 *
 * If you receive this source code without DJI’s authorization, you may not
 * further disseminate the information, and you must immediately remove the
 * source code and notify DJI of its removal. DJI reserves the right to pursue
 * legal actions against you for any loss(es) or damage(s) caused by your
 * failure to do so.
 *
 *********************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include "camera_stream_show.hpp"
#include <gst/app/gstappsrc.h>
#include <gst/app/gstappsink.h>
#include <gst/gstbuffer.h>

/* Private constants ---------------------------------------------------------*/

/* Private types -------------------------------------------------------------*/

/* Private values -------------------------------------------------------------*/

/* Private functions declaration ---------------------------------------------*/

/* Exported functions definition ---------------------------------------------*/
CameraStreamShow::CameraStreamShow()
{
    strcpy(m_currentShowImageFormat, "unknown");
    m_currentShowImageWidth = 0;
    m_currentShowImageHeight = 0;

	if (CreateShowPipeline() != ErrorCode::SUCCESS) {
		DERROR("Create show pipeline error");
		exit(1);
	}
}

CameraStreamShow::~CameraStreamShow()
{
	DestroyShowPipeline();
}

ErrorCode::ErrorCodeType CameraStreamShow::CreateShowPipeline()
{
	GstStateChangeReturn ret;

	m_gstShowAppSrc = gst_element_factory_make("appsrc", "image_show_app_src");
	m_gstShowCapsfilter = gst_element_factory_make("capsfilter", "showCapsFilter");
	m_gstShowSink = gst_element_factory_make("nveglglessink", "image_show_sink");
	m_gstShowPipeline = gst_pipeline_new("dji_video_stream_show_pipeline");

	if (!m_gstShowPipeline || !m_gstShowAppSrc || !m_gstShowSink || !m_gstShowCapsfilter) {
		DERROR("Not all gstreamer elements could be created.");
		return ErrorCode::SYSTEM_ERROR;
	}

	//appsrc config
	/* configure appsrc to be a live source */
	gst_base_src_set_live(GST_BASE_SRC (m_gstShowAppSrc), TRUE);
	/* make appsrc output a segment in time */
	gst_base_src_set_format(GST_BASE_SRC (m_gstShowAppSrc), GST_FORMAT_TIME);
	/* make appsrc set timestamps on outgoing buffers based on the running_time
	 * when they were captured */
	gst_base_src_set_do_timestamp(GST_BASE_SRC (m_gstShowAppSrc), TRUE);

	/* Build the pipeline */
	gst_bin_add_many(GST_BIN (m_gstShowPipeline), m_gstShowAppSrc, m_gstShowCapsfilter, m_gstShowSink, NULL);
	if (gst_element_link_many(m_gstShowAppSrc, m_gstShowCapsfilter, m_gstShowSink, NULL) != TRUE) {
		DERROR("Gstreamer elements could not be linked.");
		gst_object_unref(m_gstShowPipeline);
		return ErrorCode::SYSTEM_ERROR;
	}

	/* Start playing */
	ret = gst_element_set_state(m_gstShowPipeline, GST_STATE_PLAYING);
	if (ret == GST_STATE_CHANGE_FAILURE) {
		DERROR("Unable to set the pipeline to the playing state.");
		gst_object_unref(m_gstShowPipeline);
		return ErrorCode::SYSTEM_ERROR;
	}

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraStreamShow::DestroyShowPipeline()
{
    gst_element_set_state(m_gstShowPipeline, GST_STATE_PAUSED);

	gst_object_unref(m_gstShowPipeline);

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraStreamShow::RestartShowPipeline()
{
	ErrorCode::ErrorCodeType errorCodeType;
	GstStateChangeReturn stateChangeRet;

	errorCodeType = DestroyShowPipeline();
	if (errorCodeType != ErrorCode::SUCCESS) {
		DERROR("Destroy show pipeline error");
		return errorCodeType;
	}

	errorCodeType = CreateShowPipeline();
	if (errorCodeType != ErrorCode::SUCCESS) {
		DERROR("Create show pipeline error");
		return errorCodeType;
	}

	return ErrorCode::SUCCESS;
}

ErrorCode::ErrorCodeType CameraStreamShow::ShowImage(const T_Image *image)
{
	GstStateChangeReturn stateRet;
	GstBuffer *gstShowBuffer;
	GstFlowReturn ret;
	GstCaps *caps;

    if (strcmp(m_currentShowImageFormat, image->format) != 0 || m_currentShowImageWidth != image->width ||
        m_currentShowImageHeight != image->height) {
       if (RestartShowPipeline() != ErrorCode::SUCCESS) {
			DERROR("Restart show pipeline error");
			return ErrorCode::SYSTEM_ERROR;
		}
	    caps = gst_caps_new_simple("video/x-raw", "format", G_TYPE_STRING, image->format,
		       "width", G_TYPE_INT, image->width, "height", G_TYPE_INT, image->height, "framerate", GST_TYPE_FRACTION, 0, 1, nullptr);
	    g_object_set(m_gstShowCapsfilter, "caps", caps, nullptr);

	    strcpy(m_currentShowImageFormat, image->format);
	    m_currentShowImageWidth = image->width;
	    m_currentShowImageHeight = image->height;
	    DSTATUS("Change image show type, format = %s, width = %d, Height = %d", image->format, image->width, image->height);
    }

	gstShowBuffer = gst_buffer_new_allocate(nullptr, image->imageDataLen, nullptr);
	gst_buffer_fill(gstShowBuffer, 0, image->imageData, image->imageDataLen);

	ret = gst_app_src_push_buffer(GST_APP_SRC(m_gstShowAppSrc), gstShowBuffer);
	if (ret != GST_FLOW_OK) {
		DERROR("Gstreamer showAppSrc push buffer error!");
	}

	return ErrorCode::SUCCESS;
}

/****************** (C) COPYRIGHT DJI Innovations *****END OF FILE****/